!+
! Module: ekstest_mod
!
! Authors:  Natalie Webb - CESR
!          
! Description: Carries out a Kolmogorov Smirnov test of variability on EPIC data.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module ekstest_mod

  use dal
  use param  
  use errorHandling
  use metatask_mod
  use dss
  use math_utils

  implicit none 
  
  type(TaskInfo) :: info ! type from metatask_mod
 
  type test
     real(kind=single) :: Stat , Prob
  end type test

  type data_tests
     real(kind=single) :: Mean_Count, Mean_Count_bkg, Variance
     type(test) :: Chi_Square, K_S
  end type data_tests

  private 

  public ekstest

contains 
  
  !---------------------------------------------------------------------
  subroutine ekstest
  !---------------------------------------------------------------------
    
  ! Input parameters :
  !   set : Source time series FITS file name
  ! Dimension variables :
  !   Nlight : Number of light-curve contained in the FITS file (1<Nlight<9)
  !   Nbin : Number of bin in all light-curves
  ! Time variables :
  !   Time_Vect : Vector containing middle of each time interval
  !   Bin_Width
  ! Counts variables :
  !   Net_Rate, Net_Err, Bkg_Rate, Bkg_Err, Net_Counts : Vectors containing
  !    respectively net count rate & error and background count 
  !    rate & error and net source count. Their size is Nlight*Nbin
  ! Label variable : Array containing the labels to add on plot
  ! Good_Bin : Vector explaining which bins can be tested
     
    character(len=128) :: Input_File, GTI_File, Output_File
    real(kind=single) :: Bin_Width, D, KSprob, Chi_Square_Stat, Mean_rate
    real(kind=single) :: Chi_Square_Prob, sigma, sigmaflare, sigmavar, BkgRatio
    real(kind=single),dimension(:),pointer :: Time_Vect
    real(kind=single),dimension(:,:),pointer :: Net_Rate, Net_Err, Bkg_Rate
    real(kind=single),dimension(:,:),pointer :: Bkg_Err, Frac_exp
    character(len=30) :: Content
    integer :: Nlight , Nbin, Nbin_gtis
    character(len=12) :: backapp
    character(len=6) :: fcopy = 'fcopy '
    character(len=512) :: Command
    type(data_tests) :: Light_Curve

! Read the input file and its associated GTI file if it is provided

    Input_File = StringParameter("set")
 
    if (.not. booleanParameter("screen") .and. &
        .not. booleanParameter("newoutset")) call error('EKSTEST',&
        'Require new output file name')
   
    if (booleanParameter("newoutset")) then 
      Output_File = StringParameter("outset")
       Command = fcopy // ' "'// trim(Input_File) // '" '// ' "!'// &
                                                  trim(Output_File) // '"'
      call system(Command) 
      Input_File=Output_File
    end if

    call read_file(Input_File, Bin_Width, Time_Vect, Frac_exp, Net_Rate,&
          Net_Err, Bkg_Rate, Bkg_Err, Nlight, Nbin, Content, backapp, BkgRatio)

    if (booleanParameter("gtis")) then 
      GTI_File = StringParameter("gtiset")
      call readgtis(GTI_File, Nlight, Nbin, Time_Vect, Frac_exp, Net_Rate, &
                                              Bkg_Rate, Nbin_gtis)  
    end if

    call delete_gaps(Nlight, Nbin, Net_Rate, Bkg_Rate, Frac_exp) 

! Carry out the variability tests as requested  and write the results 
! into the header

    if (Nbin > 2) then
      call variability_tests(Frac_exp, Net_Rate, Bkg_Rate, Time_Vect,  &
        Net_Rate, Nlight, Nbin, D, KSprob, Chi_Square_Stat, Chi_Square_Prob, &
                         sigmaflare, sigmavar, sigma,Bin_Width, Mean_rate, BkgRatio)
       call write_results(Input_File, D, Nbin, KSprob, Nlight, &
         Chi_Square_Stat, Chi_Square_Prob, sigmaflare, sigmavar, Mean_rate)
    else
       call warning('EKSTEST','Not enough bins to process variability tests')
    end if

    deallocate(Net_Rate, Net_Err, Bkg_Rate, Bkg_Err, Time_Vect)

  end subroutine ekstest

  !-------------------------------------------------------------------------------------
  subroutine read_file(Input_File, Bin_Width, Time_Vect, Frac_exp, &
   Net_Rate, Net_Err, Bkg_Rate, Bkg_Err, Nlight, Nbin, Content, backapp, BkgRatio)
  !-------------------------------------------------------------------------------------
  ! Reads source time series FITS file and recovers all data needed. Also 
  ! creates time vector.

    character(len=*),intent(in) :: Input_File
    real(kind=single),intent(out) :: Bin_Width
    real(kind=double),dimension(:),pointer :: Array_Time
    real(kind=single),dimension(:),pointer :: Time_Vect, Time_Vect_temp
    real(kind=single),dimension(:,:),pointer :: Net_Rate, Net_Err, Bkg_Rate
    real(kind=single),dimension(:,:),pointer :: Bkg_Err,Frac_exp
    real(kind=single),dimension(:,:),pointer :: Net_Rate_temp, Net_Err_temp 
    real(kind=single),dimension(:,:),pointer :: Bkg_Rate_temp, Bkg_Err_temp
    real(kind=single),dimension(:,:),pointer :: Frac_exp_temp
    character(len=*),intent(out) :: Content
    integer,intent(out) :: Nlight, Nbin
    character(len=12) :: backapp

  ! Local variables are Set, Table, Column and Pointer needed by DAL to recover
  !  FITS binary table. Time_Zero contains the start exposure time in MJD

    type( DataSetT ) :: Ts_Set
    type( TableT ) :: Ts_Table
    type( ColumnT ) :: Col_Rate, Col_Error, Col_Backv, Col_Backe
    type( ColumnT ) :: Col_Time, Col_Frac_exp
    real(kind=single),dimension(:,:),pointer :: Array_Rate, Array_Error
    real(kind=single),dimension(:,:),pointer :: Array_Backv, Array_Backe
    real(kind=single),dimension(:,:),pointer :: Array_Frac_exp
    real(kind=single) :: Exposure_Time, Time_Zero, BkgRatio
    integer :: i, icnt, jcnt, j

    call read_data (Input_File, Ts_Set, Content, backapp)

! The keywords TSTART, TIMEDEL, and EXPOSURE TIME are in the RATE extension 
! not anymore in the primary header

    call read_table(Input_File, Ts_Set, Ts_Table, Time_Zero, Bin_Width, &
                                               Exposure_Time, BkgRatio)

    Col_Time = Column(Ts_Table,'TIME',Read)
    Col_Rate = Column(Ts_Table , ColumnNumber(Ts_Table,'RATE'),Read)
    Col_Frac_exp = Column(Ts_Table , ColumnNumber(Ts_Table,'FRACEXP'),Read)
    Col_Error = Column(Ts_Table , ColumnNumber(Ts_Table,'ERROR'),Read)
    Col_Backv = Column(Ts_Table , ColumnNumber(Ts_Table,'BACKV'),Read)
    Col_Backe = Column(Ts_Table , ColumnNumber(Ts_Table,'BACKE'),Read)

    Array_Time => real64Data(Col_Time)
    Array_Frac_exp => real32VectorData(Col_Frac_exp)
    Array_Rate => real32VectorData(Col_Rate)
    Array_Error => real32VectorData(Col_Error)
    Array_Backv => real32VectorData(Col_Backv)
    Array_Backe => real32VectorData(Col_Backe)

    allocate(Time_Vect_temp(size(Array_Time)))
    allocate(Net_Rate_temp( size(Array_Rate(:,0)) , size(Array_Rate(0,:))))
    allocate(Frac_exp_temp( size(Array_Frac_exp(:,0)),size(Array_Frac_exp(0,:))))
    allocate(Net_Err_temp( size(Array_Error(:,0)),size(Array_Error(0,:))))
    allocate(Bkg_Rate_temp( size(Array_Backv(:,0)),size(Array_Backv(0,:))))
    allocate(Bkg_Err_temp( size(Array_Backe(:,0)),size(Array_Backe(0,:))))

! You should note that although the count rates are called NET - they are in 
! fact the NON-background subtracted rates.

    if (trim(backapp) == 'F' .and. booleanParameter("twodists")) then
 
!already have the total counts necessary for this version of the KS test

      j=0

      do i=0,size(Array_Time(:))-1
        if (.not. isNull(Col_Rate,i)) then
          j=j+1
          Time_Vect_temp(j) = sngl(Array_Time(i))
          Net_Rate_temp(:,j) = Array_Rate(:,i) 
          Frac_exp_temp(:,j) = Array_Frac_exp(:,i) 
          Net_Err_temp(:,j) = Array_Error(:,i)
          Bkg_Rate_temp(:,j) = Array_Backv(:,i) 
          Bkg_Err_temp(:,j) = Array_Backe(:,i)
        end if
      end do

    allocate(Time_Vect(j))
    allocate(Net_Rate( size(Array_Rate(:,0)) , j))
    allocate(Frac_exp( size(Array_Frac_exp(:,0)) , j))
    allocate(Net_Err( size(Array_Error(:,0)) , j))
    allocate(Bkg_Rate( size(Array_Backv(:,0)) , j))
    allocate(Bkg_Err( size(Array_Backe(:,0)) , j))

    Time_Vect = Time_Vect_temp(1:j) 
    Frac_exp = Frac_exp_temp(:,1:j) 
    Net_Rate = Net_Rate_temp(:,1:j) 
    Net_Err = Net_Err_temp(:,1:j) 
    Bkg_Rate = Bkg_Rate_temp(:,1:j) 
    Bkg_Err = Bkg_Err_temp(:,1:j) 

    deallocate(Time_Vect_temp, Frac_exp_temp, Net_Rate_temp, Net_Err_temp)
    deallocate(Bkg_Err_temp, Bkg_Rate_temp)

    else if (trim(backapp) == 'T' .and. (.not. booleanParameter("twodists"))) then

!want net counts only in this situation

      j=0

      do i=0,size(Array_Time(:))-1
        if (.not. isNull(Col_Rate,i)) then
          j=j+1
          Time_Vect_temp(j) = sngl(Array_Time(i))
          Frac_exp_temp(:,j) = Array_Frac_exp(:,i) 
          Net_Rate_temp(:,j) = Array_Rate(:,i) 
          Net_Err_temp(:,j) = Array_Error(:,i)
          Bkg_Rate_temp(:,j) = Array_Backv(:,i) 
          Bkg_Err_temp(:,j) = Array_Backe(:,i)
         end if
      end do
      allocate(Time_Vect(j))
      allocate(Net_Rate( size(Array_Rate(:,0)) , j))
      allocate(Frac_exp( size(Array_Frac_exp(:,0)) , j))
      allocate(Net_Err( size(Array_Error(:,0)) , j))
      allocate(Bkg_Rate( size(Array_Backv(:,0)) , j))
      allocate(Bkg_Err( size(Array_Backe(:,0)) , j))

      Time_Vect = Time_Vect_temp(1:j) 
      Net_Rate = Net_Rate_temp(:,1:j) 
      Frac_exp = Frac_exp_temp(:,1:j) 
      Net_Err = Net_Err_temp(:,1:j) 
      Bkg_Rate = Bkg_Rate_temp(:,1:j) 
      Bkg_Err = Bkg_Err_temp(:,1:j) 

     deallocate(Time_Vect_temp, Frac_exp_temp, Net_Rate_temp, Net_Err_temp)
     deallocate(Bkg_Err_temp, Bkg_Rate_temp)

    else if (trim(backapp) == 'T' .and. booleanParameter("twodists")) then
      j=0
      do i=0,size(Array_Time(:))-1

!get the total counts

        if (.not. isNull(Col_Rate,i)) then
         j=j+1
          Time_Vect_temp(j) = sngl(Array_Time(i))
          Net_Rate_temp(:,j) = Array_Rate(:,i) +Array_Backv(:,i)

!This is negtive because the net counts are correlated with the total and the
!background counts

          Net_Err_temp(:,j) = sqrt(Array_Error(:,i)**2.0-Array_Backe(:,i)**2.0)
          Bkg_Rate_temp(:,j) = Array_Backv(:,i) 
          Bkg_Err_temp(:,j) = Array_Backe(:,i)
 
         end if
      end do
      allocate(Time_Vect(j))
      allocate(Net_Rate( size(Array_Rate(:,0)) , j))
      allocate(Frac_exp( size(Array_Frac_exp(:,0)) , j))
      allocate(Net_Err( size(Array_Error(:,0)) , j))
      allocate(Bkg_Rate( size(Array_Backv(:,0)) , j))
      allocate(Bkg_Err( size(Array_Backe(:,0)) , j))

      Time_Vect = Time_Vect_temp(1:j) 
      Net_Rate = Net_Rate_temp(:,1:j)
      Frac_exp = Frac_exp_temp(:,1:j) 
      Net_Err = Net_Err_temp(:,1:j) 
      Bkg_Rate = Bkg_Rate_temp(:,1:j) 
      Bkg_Err = Bkg_Err_temp(:,1:j) 

     deallocate(Time_Vect_temp, Frac_exp_temp, Net_Rate_temp, Net_Err_temp)
     deallocate(Bkg_Err_temp, Bkg_Rate_temp)
    else 
      j=0
      do i=0,size(Array_Time(:))-1

!get the net counts

        if (.not. isNull(Col_Rate,i)) then
         j=j+1
          Time_Vect_temp(j) = sngl(Array_Time(i))
          Frac_exp_temp(:,j) = Array_Frac_exp(:,i) 
          Net_Rate_temp(:,j) = Array_Rate(:,i)-Array_Backv(:,i)
          Net_Err_temp(:,j) = sqrt(Array_Error(:,i)**2.0+Array_Backe(:,i)**2.0)
          Bkg_Rate_temp(:,j) = Array_Backv(:,i) 
          Bkg_Err_temp(:,j) = Array_Backe(:,i)
 
         end if
      end do
      allocate(Time_Vect(j))
      allocate(Net_Rate( size(Array_Rate(:,0)) , j))
      allocate(Frac_exp( size(Array_Frac_exp(:,0)) , j))
      allocate(Net_Err( size(Array_Error(:,0)) , j))
      allocate(Bkg_Rate( size(Array_Backv(:,0)) , j))
      allocate(Bkg_Err( size(Array_Backe(:,0)) , j))

      Time_Vect = Time_Vect_temp(1:j) 
      Net_Rate = Net_Rate_temp(:,1:j) 
      Frac_exp = Frac_exp_temp(:,1:j) 
      Net_Err = Net_Err_temp(:,1:j) 
      Bkg_Rate = Bkg_Rate_temp(:,1:j) 
      Bkg_Err = Bkg_Err_temp(:,1:j) 

     deallocate(Time_Vect_temp, Frac_exp_temp, Net_Rate_temp, Net_Err_temp)
     deallocate(Bkg_Err_temp, Bkg_Rate_temp)
    end if

!    do jcnt=1,size(Net_Rate(:,0))   
!      do icnt=1,size(Net_Rate(0,:))   
    do jcnt=1,size(Array_Rate(:,0)) 
      do icnt=1,j 
       if ((trim(backapp) == 'F') .and. (booleanParameter("twodists")) .and. (Net_Rate(jcnt,icnt) < 0.0d0)) call &
          error('EKSTEST','Negative count rates in '//Input_File)
       if ((trim(backapp) == 'T') .and. (booleanParameter("twodists")) &
         .and. (Net_Rate(jcnt,icnt) < 0.0d0)) call error('EKSTEST',&
        'Negative count rates in '//Input_File)
       if (Bkg_Rate(jcnt,icnt) < 0.0d0) call error('EKSTEST', &
        'Negative background values in '//Input_File)
      end do
    end do

    call check_sizes (Frac_exp, Net_Rate, Net_Err, Bkg_Rate, Bkg_Err, Nlight, Nbin)

    call release(Col_Time)
    call release(Col_Rate)
    call release(Col_Frac_exp)
    call release(Col_Error)
    call release(Col_Backv)
    call release(Col_Backe)
    call release(Ts_Table)    
    call release(Ts_Set)

    ! More checks

    call check_sizes (Frac_exp, Net_Rate, Net_Err, Bkg_Rate, Bkg_Err, Nlight, Nbin)

!    allocate( Time_Vect(Nbin) )
!    do i=1,Nbin
!       Time_Vect(i) = (i-0.5)*Bin_Width
!    end do

  end subroutine read_file

  !-------------------------------------------------------------------------------------
  subroutine read_data (Input_File , Ts_Set, Content, backapp)
  !-------------------------------------------------------------------------------------
  ! Reads the in parameter FITS file and gives back a handle to 
  ! corresponding dataset. Checks some keywords existence and validity.

    character(len=*),intent(in) :: Input_File 
    type( DataSetT ),intent(out) :: Ts_Set
    character(len=*),intent(out) :: Content
    character(len=12),intent(out) :: BACKAPP

    type( TableT ) :: Table1

    Ts_Set = dataSet(Input_File,READ)

    Table1 = table(Ts_Set,BlockNumber(Ts_Set,'RATE')) 

    if (HasAttribute(Ts_Set,"BACKAPP") ) then
         backapp = StringAttribute(Ts_Set,'BACKAPP')
    else if (HasAttribute(Table1,"BACKAPP") ) then
         backapp = StringAttribute(Table1,'BACKAPP')
    else
         backapp = 'F'
    endif

    if ( .not. HasAttribute(Ts_Set,"CONTENT") ) then
!      call warning('EKSTEST','Keyword CONTENT missing '//Input_File)
        Content = 'SOURCE TIMESERIES'
    else
        Content = change_case(StringAttribute(Ts_Set,"CONTENT"),2)
!       if ( ( Content /= 'EPIC SOURCE TIMESERIES' ) .and. &
!         ( Content /='OM OSW SOURCE TIMESERIES' ) ) &
!    call warning('EKSTEST','CONTENT parameter for input file '//Input_File// &
!                          ' is : '//Content)
    endif


  end subroutine read_data

  !-------------------------------------------------------------------------------------
  subroutine read_table (Input_File, Ts_Set, Ts_Table, Time_Zero, Bin_Width, &
       Exposure_Time, BkgRatio )
  !-------------------------------------------------------------------------------------
  ! With FITS file name and handle to dataset, generates table.
  ! Checks more keywords.
  ! Also gives back Time_Zero, Bin_Width, Exposure_Time and Content.

    character(len=*),intent(in) :: Input_File
    type( DataSetT ),intent(in) :: Ts_Set
    type( TableT ),intent(out) :: Ts_Table

    real(kind=single),intent(out) :: Time_Zero, Bin_Width , Exposure_Time, BkgRatio

    Ts_Table = table(Ts_Set,BlockNumber(Ts_Set,'RATE'))
    if ( .not. HasAttribute(Ts_Table,"HDUCLASS") ) &
      call warning('EKSTEST','Keyword HDUCLASS missing in '//Input_File)
    if ( change_case(StringAttribute(Ts_Table,"HDUCLASS"),2) /= 'OGIP' ) &
      call warning('EKSTEST','HDUCLAS is not equal to OGIP')
    if ( .not. HasAttribute(Ts_Table,"HDUCLAS1") ) &
      call warning('EKSTEST','Keyword HDUCLAS1 missing in '//Input_File)
    if ( change_case(StringAttribute(Ts_Table,"HDUCLAS1"),2) /= 'LIGHTCURVE' )&
      call warning('EKSTEST','HDUCLAS1 is not equal to LIGHTCURVE')
!    if ( .not. HasAttribute(Ts_Table,"HDUCLAS2") ) &
!      call warning('EKSTEST','Keyword HDUCLAS2 missing in '//Input_File)
!    if ( change_case(StringAttribute(Ts_Table,"HDUCLAS2"),2) /= 'TOTAL' ) &
!      call warning('EKSTEST','HDUCLAS2 is not equal to TOTAL')
    if ( .not. HasAttribute(Ts_Table,"HDUCLAS3")) &
      call warning('EKSTEST','Keyword HDUCLAS3 missing in '//Input_File)
    if ( change_case(StringAttribute(Ts_Table,"HDUCLAS3"),2) /= 'RATE' ) &
      call warning('EKSTEST','HDUCLAS3 is not equal to RATE')

    if ( .not. HasAttribute(Ts_Table,"TSTART") ) then
      call warning('EKSTEST','Keyword TSTART missing in '//Input_File)
    else
      Time_Zero = real32Attribute(Ts_Table,'TSTART')
    endif

    if ( .not. HasAttribute(Ts_Table,"TIMEDEL") ) &
      call error('EKSTEST','Keyword TIMEDEL missing in '//Input_File)
    Bin_Width = real32Attribute(Ts_Table,'TIMEDEL')
    if ( Bin_Width < epsilon(0.) ) call error('EKSTEST', &
      'Null or negative Bin_Width in '//Input_File)

    if ( .not. HasAttribute(Ts_Table,"EXPOSURE") ) then
         call warning('EKSTEST','Keyword EXPOSURE missing in '//Input_File)
    else
         Exposure_Time = real32Attribute(Ts_Table,"EXPOSURE") 
         if ( Exposure_Time < epsilon(0.) ) call warning('EKSTEST', &
          'Null or negative exposure time in '//Input_File)
    endif

    if ( .not. HasAttribute(Ts_Table,"BKGRATIO") ) then
         call warning('EKSTEST','Keyword BKGRATIO missing in '//Input_File)
    else
         BkgRatio = real32Attribute(Ts_Table,"BKGRATIO") 
    endif

  end subroutine read_table


  !-------------------------------------------------------------------------------------
  subroutine check_sizes(Frac_exp, Net_Rate, Net_Err, Bkg_Rate, Bkg_Err, Nlight, Nbin)
  !------------------------------------------------------------------------------------
  ! Checks if all these vectors have same dimensions : Nlight * Nbin.

    real(kind=single),dimension(:,:),pointer :: Net_Rate, Net_Err, Bkg_Rate
    real(kind=single),dimension(:,:),pointer :: Bkg_Err, Frac_exp
    integer,intent(out) :: Nlight, Nbin

    Nlight = size(Net_Rate(:,1))
    if ( size(Frac_exp(:,1)) /= Nlight ) call error('EKSTEST', &
       'Different number of columns between EXP, RATE, ERROR, BACKV and BACKE')
    if ( size(Net_Err(:,1)) /= Nlight ) call error('EKSTEST', &
       'Different number of columns between EXP, RATE, ERROR, BACKV and BACKE')
    if ( size(Bkg_Rate(:,1)) /= Nlight ) call error('EKSTEST', &
       'Different number of columns between EXP, RATE, ERROR, BACKV and BACKE')
    if ( size(Bkg_Err(:,1)) /= Nlight ) call error('EKSTEST', &
       'Different number of columns between EXP, RATE, ERROR, BACKV and BACKE')
    if ( Nlight > 9 ) call error('EKSTEST', &
         'Number of light-curves contained in a file must not exceed 9')
    if (.not. ( size(Net_Rate(1,:)) == size(Net_Err(1,:)) &
     .and. size(Bkg_Rate(1,:)) == size(Bkg_Err(1,:)) &
     .and. size(Frac_exp(1,:)) == size(Net_Rate(1,:)) &
     .and. size(Net_Rate(1,:)) == size(Bkg_Rate(1,:)) )) call error('EKSTEST',&
       'Different number of rows in columns EXP, RATE, ERROR, BACKV and BACKE')
    Nbin = size( Net_Rate(1,:) )
    if ( Nbin <= 0 ) call error('EKSTEST','No data point received')

  end subroutine check_sizes 


!-------------------------------------------------------------------------------------
  subroutine readgtis(GTI_File, Nlight, Nbin, Time_Vect, Frac_exp, Net_Rate, &
                     Bkg_Rate, Nbin_gtis) 
  !-------------------------------------------------------------------------------------

! This subroutine reads the GTI file and deletes all gaps and null 
! values in theinput data and its associated background file

    real(kind=single),dimension(:,:),pointer :: Net_Rate, Bkg_Rate,Frac_exp
    real(kind=single),dimension(:,:),pointer :: Net_Rate_temp, Bkg_Rate_temp
    real(kind=single),dimension(:,:),pointer :: Frac_exp_temp
    character(len=*),intent(in) :: GTI_File
    integer,intent(inout) :: Nbin, Nlight

  ! Local variables are Set, Table, Column and Pointer needed by DAL to 
  ! recover FITS binary table. Moreover Time_Zero contains the start 
  ! exposure time in MJD

    type( DataSetT ) :: Gti_Set
    type( TableT ) :: Gti_Table
    type( ColumnT ) :: Col_Start, Col_Stop
    real(kind=double),dimension(:),pointer :: Array_Start, Array_Stop
    real(kind=single),dimension(:),pointer :: Time_Vect, Time_Vect_temp
    real(kind=single) :: T_Start, T_Stop, time_exp
    integer :: icnt, jcnt, kcnt, numofgtis, Nbin_gtis
    real(kind=single),dimension(:),allocatable  :: listopt
    real :: Eps

    Gti_Set = dataSet(GTI_File,READ)

    Gti_Table = table(Gti_Set,BlockNumber(Gti_Set,'STDGTI')) 

    Col_Start = Column(Gti_Table, ColumnNumber(Gti_Table,'START'),Read)
    Col_Stop = Column(Gti_Table, ColumnNumber(Gti_Table,'STOP'),Read)

   if ( .not. HasAttribute(Gti_Table,"TSTART") ) then
      call warning('EKSTEST','Keyword TSTART missing in '//GTI_File)
    else
      T_Start = real32Attribute(Gti_Table,'TSTART')
    endif

   if ( .not. HasAttribute(Gti_Table,"TSTOP") ) then
      call warning('EKSTEST','Keyword TSTOP missing in '//GTI_File)
    else
      T_Stop = real32Attribute(Gti_Table,'TSTOP')
    endif

    Array_Start => real64Data(Col_Start)
    Array_Stop => real64Data(Col_Stop)

! This is the exposure time for the input timeseries

    numofgtis = size(Array_Start)
    time_exp = Time_Vect(Nbin) - Time_Vect(1)
!    Nbin_gtis=int(time_exp)
    Nbin_gtis=Nbin

    allocate(listopt((numofgtis*2)))

    jcnt=0
    do icnt=0,numofgtis-1
      listopt(jcnt+1)=((Array_Start(icnt)/86400.0)+10814.0)
      listopt(jcnt+2)=((Array_Stop(icnt)/86400.0)+10814.0)
      jcnt=jcnt+2
    end do

!    do icnt=1,jcnt
!      write(10,*)listopt(icnt)
!    end do
    
! Now get rid of any gaps or null values in the background

    allocate(Net_Rate_temp(Nlight,Nbin), Bkg_Rate_temp(Nlight,Nbin), &
              Time_Vect_temp(Nbin),Frac_exp_temp(Nlight,Nbin))

    Eps = epsilon(0.)
    kcnt=0
    do icnt=1,Nbin
      do jcnt=0,numofgtis-1
        if (Time_Vect(icnt) > Time_Vect(Nbin)) exit
!        if ((Time_Vect(icnt) >= Array_Start(jcnt)-T_Start) .and. &
!              (Time_Vect(icnt) <= Array_Stop(jcnt)-T_Start) .and. &
!              (Bkg_Rate(1,icnt) > Eps)) then
        if ((Time_Vect(icnt) >= Array_Start(jcnt)) .and. &
              (Time_Vect(icnt) < Array_Stop(jcnt)) .and. &
              (Bkg_Rate(1,icnt) >= 0.0d0)) then
         kcnt=kcnt+1
         Frac_exp_temp(1,kcnt) = Frac_exp(1,icnt) 
         Net_Rate_temp(1,kcnt) = Net_Rate(1,icnt) 
         Bkg_Rate_temp(1,kcnt) = Bkg_Rate(1,icnt) 
         Time_Vect_temp(kcnt) = Time_Vect(icnt)
        end if
      end do
    end do
    
    Nbin=kcnt
    deallocate(Net_Rate, Bkg_Rate, Time_Vect,Frac_exp)

!At the moment this file only carries out one var test on the first file. If we
!find that ekstest needs to read more files, this can be expanded.

    Nlight=1
    allocate(Net_Rate(Nlight,Nbin), Bkg_Rate(Nlight,Nbin), Time_Vect(Nbin), &
                     Frac_exp(Nlight,Nbin))

    Frac_exp(Nlight,:) =  Frac_exp_temp(1,1:Nbin)
    Net_Rate(Nlight,:) =  Net_Rate_temp(1,1:Nbin)
    Bkg_Rate(Nlight,:) =  Bkg_Rate_temp(1,1:Nbin)
    Time_Vect(:) = Time_Vect_temp(1:Nbin)

    deallocate(Net_Rate_temp, Bkg_Rate_temp, Time_Vect_temp, Frac_exp_temp)

    call release(Col_Start)
    call release(Col_Stop)
    call release(Gti_Table)
    call release(Gti_Set)

  end subroutine readgtis


!-------------------------------------------------------------------------------------
  subroutine delete_gaps(Nlight, Nbin, Net_Rate, Bkg_Rate, Frac_exp) 
  !-------------------------------------------------------------------------------------

! This subroutine reads the GTI file and deletes all gaps and null values 
! in the input data and its associated background file and the FRACEXP

    real(kind=single),dimension(:,:),pointer :: Net_Rate, Bkg_Rate,Frac_exp
    real(kind=single),dimension(:,:),pointer :: Net_Rate_temp, Bkg_Rate_temp
    real(kind=single),dimension(:,:),pointer :: Frac_exp_temp
    integer,intent(inout) :: Nbin, Nlight
    integer :: icnt, kcnt
    real :: Eps

! Now get rid of any gaps or null values in background or if the FRACEXP is zero

    allocate(Net_Rate_temp(Nlight,Nbin), Bkg_Rate_temp(Nlight,Nbin))
    allocate(Frac_exp_temp(Nlight,Nbin))

    Eps = epsilon(0.)
    kcnt=0
    do icnt=1,Nbin
!        if ((Net_Rate(1,icnt) > Eps) .and. (Bkg_Rate(1,icnt) > Eps)) then
        if (Bkg_Rate(1,icnt) >= 0.0d0  .or.  Frac_exp(1,icnt) == 0.0) then
         kcnt=kcnt+1
         Frac_exp_temp(1,kcnt) = Frac_exp(1,icnt) 
         Net_Rate_temp(1,kcnt) = Net_Rate(1,icnt) 
         Bkg_Rate_temp(1,kcnt) = Bkg_Rate(1,icnt) 
        end if
    end do
    
    Nbin=kcnt
    deallocate(Net_Rate, Bkg_Rate, Frac_exp)

!At the moment this file only carries out one KS test on the first file.  If we
!find that ekstest needs to read more files, this can be expanded.

    Nlight=1
    allocate(Net_Rate(Nlight,Nbin), Bkg_Rate(Nlight,Nbin), Frac_exp(Nlight,Nbin))

    Net_Rate(Nlight,:) =  Net_Rate_temp(1,1:Nbin)
    Bkg_Rate(Nlight,:) =  Bkg_Rate_temp(1,1:Nbin)
    Frac_exp(Nlight,:) =  Frac_exp_temp(1,1:Nbin)

    deallocate(Net_Rate_temp, Bkg_Rate_temp, Frac_exp_temp)


  end subroutine delete_gaps


  !-------------------------------------------------------------------------------------
  subroutine variability_tests(Frac_exp, Net_Rate, BCount, Time_Vect, &
                              Net_total, Nlight, Nbin, D, KSprob, &
                              Chi_Square_Stat, Chi_Square_Prob, sigmaflare, &
                              sigmavar, sigma, bin_del, avrate, bkgRatio)
  !-------------------------------------------------------------------------------------
! Adds up the rate values and their associated background rate (called BCount),
! carries out variance tests, chi-squared tests and kolmogorov smirnov tests

    real(kind=single),dimension(:,:),pointer :: Frac_exp, Net_Rate, Net_total
    real(kind=single),dimension(:,:),pointer :: BCount
    real(kind=single),dimension(:),pointer :: Time_Vect
    integer,intent(in) :: Nlight, Nbin
    real(kind=single):: D, KSprob, Chi_Square_Stat, Chi_Square_Prob, bin_del
    real(kind=single):: avrate, BkgRatio

    type(data_tests),dimension(Nlight) :: Light_Curve, Light_Curve_bkg
    type(data_tests),dimension(Nlight) :: Light_Curve_total
    real :: sigma, sigmaflare, sigmavar
    integer :: Light, i

    do Light=1,Nlight
       Light_Curve(Light)%Mean_Count = sum(Net_Rate(Light,:)) / Nbin
       Light_Curve_total(Light)%Mean_Count = sum(Net_total(Light,:)) / Nbin
       Light_Curve_bkg(Light)%Mean_Count_bkg = sum(BCount(Light,:)) / Nbin
       Light_Curve(Light)%Variance = 0.
       Light_Curve_total(Light)%Variance = 0.
       do i=1,Nbin
         Light_Curve(Light)%Variance = Light_Curve(Light)%Variance + &
               (Net_Rate(Light,i) - Light_Curve(Light)%Mean_Count)**2.0
         Light_Curve_total(Light)%Variance = Light_Curve_total(Light)%Variance&
             + (Net_total(Light,i) - Light_Curve_total(Light)%Mean_Count)**2.0
       end do
       Light_Curve(Light)%Variance = Light_Curve(Light)%Variance / (Nbin-1)
       Light_Curve_total(Light)%Variance = Light_Curve_total(Light)%Variance &
                                                               / (Nbin-1)
       sigma=sqrt(Light_Curve(Light)%Variance)

!       call kolmogorov_smirnov_test( Net_total(Light,:) , Nbin , &
!            Light_Curve_total(Light) )

     if (booleanParameter("chisquaretest") .and. booleanParameter("twodists"))&
        call error('EKSTEST','Can not carry out two distribution KS test and chi-squared test')   

     if (booleanParameter("chisquaretest") .and. &
           (.not. booleanParameter("twodists"))) then

       call chi_square_test(Frac_exp(Light,:), Net_Rate(Light,:) , BCount(Light,:), Nbin , &
            Light_Curve(Light), bin_del, avrate, BkgRatio)
       Chi_Square_Stat=Light_Curve(Light)%Chi_Square%Stat
       Chi_Square_Prob=Light_Curve(Light)%Chi_Square%Prob
     end if

     if (booleanParameter("kstest")) then
       if (booleanParameter("twodists")) then
         call kolmogorov_smirnov_test2(Net_Rate(Light,:), BCount(Light,:), &
            Nbin, Light_Curve(Light),  Light_Curve_bkg(Light), D, KSprob)
       else 
!        Net_Rate = Net_Rate-BCount
         call kolmogorov_smirnov_test2(Net_Rate(Light,:), Time_Vect, Nbin, &
            Light_Curve(Light),  Light_Curve_bkg(Light), D, KSprob)
       end if
     end if

     if (booleanParameter("flaretest") .and. booleanParameter("twodists"))&
        call error('EKSTEST','Can not carry out two distribution KS test and chi-squared test')   

     if (booleanParameter("flaretest") .and. &
              (.not. booleanParameter("twodists"))) then
       call flare_test(Net_Rate(Light,:), Nbin , sigma, sigmaflare)
     end if

     if (booleanParameter("variationtest") .and. booleanParameter("twodists"))&
        call error('EKSTEST','Can not carry out two distribution KS test and chi-squared test')   

     if (booleanParameter("variationtest") .and. &
              (.not. booleanParameter("twodists"))) then
       call var_test(Net_Rate(Light,:), Nbin, sigma, sigmavar)
     end if

    end do


  end subroutine variability_tests

  !-------------------------------------------------------------------------------------
  subroutine write_results(Input_File, D, Nbin, KSprob, Nlight, &
                Chi_Square_Stat, Chi_Square_Prob, sigmaflare, sigmavar, avrate)
  !-------------------------------------------------------------------------------------

! This routine writes the results into the header

    character(len=*),intent(in) :: Input_File 
    type( DataSetT ) :: set
    type( TableT ) :: tab
    real(kind=single) :: D, KSprob, Chi_Square_Stat, Chi_Square_Prob
    real(kind=single) :: sigmaflare, sigmavar, avrate
    integer :: Nbin, Nlight
    type(data_tests),dimension(Nlight) :: Light_Curve

    set = dataSet(Input_File, MODIFY)
    tab = table(set, BlockNumber(set,'RATE'))


    if (booleanParameter("screen")) then 
      if (booleanParameter("kstest")) then
        print *, '          ksStatistic=', D
        print *, '        ksProbability=', KSprob
      end if
      if (booleanParameter("chisquaretest")) then
        print *, '          lc average rate   =', avrate
        print *, '          ChiSquareStatistic=', Chi_Square_Stat
        print *, '        ChiSquareProbability of constancy=', Chi_Square_Prob
      end if
      if (booleanParameter("flaretest")) then
        print *, '        PresenceOfFlare=', sigmaflare
      end if
      if (booleanParameter("variationtest")) then
        print *, '        PresenceOfVariation=', sigmavar
      end if
!print *, '      kuiperStatistic=', kuiperStatistic
!print *, '    kuiperProbability=', kuiperProbability
!print *, '  chiSquaredStatistic=', chiSquaredStatistic
!print *, 'chiSquaredProbability=', chiSquaredProbability
       print *, '          numNonNulls=', Nbin
    end if

    if (booleanParameter("newoutset")) then

      call setAttribute(tab, 'N_POINTS', Nbin &
      , comment='The number of data points.')

      
      if (booleanParameter("kstest")) then
        call setAttribute(tab, 'KVSVSTAT', D &
        , comment='Kolmogorov-Smirnov statistic.')
        call setAttribute(tab, 'KVSVPROB', KSprob  &
        , comment='Probability that KVSVSTAT arises by chance')
      end if
      if (booleanParameter("chisquaretest")) then
      call setAttribute(tab, 'AVRATE', avrate &
      , comment='Average rate in light curve')
      call setAttribute(tab, 'CHISQUAR', Chi_Square_Stat &
      , comment='Chi squared value.')
      call setAttribute(tab, 'CHI2PROB', Chi_Square_Prob &
      , comment='Probability that CHISQUAR arises by chance')
       end if
      if (booleanParameter("flaretest")) then
        call setAttribute(tab, 'FLARE', sigmaflare &
        , comment='Peak in data (no. of sigma above median)') 
      end if   
      if (booleanParameter("variationtest")) then
        call setAttribute(tab, 'VARIATIO', sigmavar &
        , comment='Variation (sigma) between two ends of data') 
      end if   
!      call setAttribute(tab, 'KUIPSTAT', kuiperStatistic&
!      , comment="Kuiper ('cyclic Kolmogorov-Smirnov') statistic.")
!      call setAttribute(tab, 'KUIPPROB', kuiperProbability&
!      , comment='Probability that KUIPSTAT value would arise by chance.')
!      call setAttribute(tab, 'N_EFF', numEffective&
!      , comment='The effective number of data points, used to calculate KS &
!      &and Kuiper probabilities.')
!      call setAttribute(tab, 'CHISQUAR', chiSquaredStatistic&
!      , comment='Chi squared value.')
!      call setAttribute(tab, 'REDCHISQ', chiSquaredStatistic/real(numNonNulls-1)&
!      , comment='Reduced chi squared value.')
!      call setAttribute(tab, 'CHI2PROB', chiSquaredProbability&
!      , comment='Probability that CHISQUAR value would arise by chance.')
 !     call setAttribute(tab, 'DEG_FREE', degFree&
!      , comment='Chi squared degrees of freedom.')

     end if

    call release(set)

end subroutine write_results

 !-------------------------------------------------------------------------------------
  subroutine flare_test(Counts, Nbin, sigma, sigmaflare)
  !-------------------------------------------------------------------------------------
  ! Returns the average sigma above the median of the flare, which has the 
  ! number of bins (flarebins) and minimum sigma (flarefactor)
  ! above the mean specified by the user.

    real(kind=single),dimension(:) :: Counts
    integer,intent(in) :: Nbin
    real(kind=single) :: median_value, sigma
    real(kind=single) :: sigmaflare, sigmaflaremax, CountsInFlare
    integer :: icnt, jcnt, factor, binmin

    call median(Counts, Nbin, median_value)

    factor = IntParameter("flarefactor")
    binmin = IntParameter("flarebins")

    jcnt=0; icnt=0
    sigmaflare = 0.0
    sigmaflaremax = 0.0

    do while (icnt < Nbin-binmin-1)
      icnt=icnt+1
      if (Counts(icnt) >= (factor*sigma)+median_value) then
        CountsInFlare=0.0
        jloop: do jcnt=0,binmin-1
          if (Counts(icnt+jcnt) < (factor*sigma)+median_value) exit jloop
          CountsInFlare=CountsInFlare+Counts(icnt+jcnt)
          if (jcnt == binmin-1) then
            sigmaflare = ((CountsInFlare/real(binmin))-median_value)/sigma
            if (sigmaflare > sigmaflaremax) sigmaflaremax = sigmaflare
          end if
        end do jloop
        jcnt=jcnt-1
      end if 
    end do

    sigmaflare=sigmaflaremax

  end subroutine flare_test

 !-------------------------------------------------------------------------------------
  subroutine var_test(Counts, Nbin, sigma, sigmavar )
  !-------------------------------------------------------------------------------------
  ! Returns the difference between the beginning and the end of the
  ! lightcurve as a sigma value

    real(kind=single),dimension(:) :: Counts
    integer,intent(in) :: Nbin
    real(kind=single) :: meanofbeg, meanofend, sigma, sigmavar
    integer :: icnt, pcent, limits

    pcent = IntParameter("percentage")

    limits=nint((real(pcent)/100.0)*Nbin)

    meanofbeg=0.0
    do icnt=1,limits
      meanofbeg=meanofbeg+Counts(icnt)
    end do

    meanofbeg=meanofbeg/real(Nbin)

    meanofend=0.0
    do icnt=Nbin,Nbin-limits,-1
      meanofend=meanofend+Counts(icnt)
    end do

    meanofend=meanofend/real(Nbin)

    sigmavar=(abs(meanofbeg-meanofend)/sigma)

  end subroutine var_test

 !-------------------------------------------------------------------------------------
  subroutine chi_square_test(FExp, Counts, BkgCounts, Nbin, Lcurve, bin_del, &
                            avrate, BkgRatio)
  !-------------------------------------------------------------------------------------
  ! Returns the chi square statistic and probability of the hypothesis 
  ! non variable source. So, the Degree of freedom (Df) is equal to Nbin-1.

    real(kind=single),dimension(:) :: Counts, BkgCounts
    real(kind=single),dimension(:) :: FExp
    integer,intent(in) :: Nbin
    type(data_tests),intent(inout) :: Lcurve

    real(kind=single) :: Stat, Mean , Rsrc, Model, var, BkgRatio, SumCount
    real(kind=single) :: bin_del, tmp, avrate, Prob_ssclib, sumback, U_i
    integer :: Df , i, status
    Df = Nbin - 1
    Stat = 0.

! BkgRatio is background area over source area, so it should be 1 over the ratio

!   Calculate weighted mean in lc  
!   We weight with the exposure values to avoid points with low exposure 
!   to affect the value of the mean

    Mean=0.0
    tmp=0.0
    sumback = 0.0
    var = 0.0
    SumCount = 0.0
    do i=1,Nbin
       Mean=Mean+Counts(i)*FExp(i)
       tmp=tmp+FExp(i)
       SumCount = SumCount + ((Counts(i) + (BkgCounts(i)/BkgRatio))*bin_del*FExp(i) - (BkgCounts(i)/BkgRatio)*bin_del*FExp(i))
       sumback = sumback + ((BkgRatio**(-2.0))*BkgCounts(i))*bin_del*FExp(i)
    end do
    Mean=Mean/tmp

    avrate=Mean

!   Provided by Ian Stewart (12/06/06)
!         <        A_s        >    A_s
!  U'_i = < u_j - ----- * b_j > + ----- * b_i        
!         <        A_b        >    A_b


!   Provided by Ian Stewart (12/06/06)
!                 ( A_s )^2
!  var_i = U'_i + (-----)  * b_i +
!                 ( A_b )

!             1               ( A_s )^2
!         + ----- sum_{j=1}^N (-----)  * b_j.
!            N^2              ( A_b )

    do i=1,Nbin
!      Calculate source raw counts(before lccor corrections)
       Rsrc=(Counts(i)+(BkgCounts(i)/BkgRatio))*bin_del*FExp(i)
       U_i = (SumCount/real(Nbin) +  (BkgCounts(i)/BkgRatio)*bin_del*FExp(i))
!      Define model
!       Model=(Mean+BkgCounts(i))*bin_del*FExp(i)

! Mode to be used to take into account of the background variability May 2006
    

       var = U_i + (BkgRatio**(-2.0))*BkgCounts(i)*bin_del*FExp(i) + (real(Nbin)**(-2.0))*sumback

!      if model=0 then do not use the data 
!       if (Model.ne.0) then
!          Stat = Stat + (Rsrc-Model)**2/abs(Model)
       if (var .ne. 0.0) then
!          Stat = Stat + (Rsrc-Model)**2/abs(Model)
          Stat = Stat + (((Rsrc-U_i)**2.0)/abs(var))
       else 
          Df=Df-1
       end if
!    call incompleteGammaQ(0.5*Df, 0.5*Stat, Prob_ssclib, status)
!    print*,Df,gamma_function_Q(0.5*Df , 0.5*Stat), Prob_ssclib
    end do

    call incompleteGammaQ(0.5*Df, 0.5*Stat, Prob_ssclib, status)
    Lcurve%Chi_Square%Stat = Stat
!    Lcurve%Chi_Square%Prob = gamma_function_Q(0.5*Df , 0.5*Stat)

! Following  	SSC-SPR-3591 and tests carried out by AS and NAW it appears 
! that the incommplete gamma function available in the ssclib gives more 
! consistent results than the gamma_function_Q (given below) and thus is 
! now used here (07/03/06)

    Lcurve%Chi_Square%Prob = Prob_ssclib

  end subroutine chi_square_test

  !-------------------------------------------------------------------------------------
  subroutine kolmogorov_smirnov_test2(Counts, BCounts, Nbin, Lcurve,&
                                                Lcurve_bkg, D, KSprob)
  !-------------------------------------------------------------------------------------
  ! Compares the cumulative rate distribution with the cumulative background
  ! distribution. Returns the corresponding K_S statistic D, and the signifiance level
  ! Prob. Great values of Prob indicate that there is no variability in the input
  ! distribution.

    real(kind=single),dimension(:) :: Counts, BCounts
    integer,intent(in) :: Nbin
    type(data_tests),intent(inout) :: Lcurve, Lcurve_bkg

    real(kind=single) :: D, en, Nbin_Real, ir, Cumulative_Poisson_Value
    real(kind=single) :: Cumulative_Poisson_Value_bkg, KSprob, fn1, fn2, f01, f02
    integer :: i, j

    D = 0.0; f01=0.0; f02=0.0; i=1; j=1
    call sort(Counts)
    call sort(BCounts)
    Nbin_Real = real(Nbin)
    en=Nbin_Real
    do while (i <= Nbin_Real .and. j <= Nbin_Real)
      if (Counts(i) < BCounts(j)) then
        fn1=i/en  
        ir=amax1(abs(fn1-f02),abs(f01-f02))
        if (ir > D) D=ir
        f01=fn1
        i=i+1
      else
        fn2=j/en  
        ir=amax1(abs(fn2-f01),abs(f02-f01))
        if (ir > D) D=ir
        f02=fn2
        j=j+1
      end if
    end do
   Lcurve%K_S%Stat = D
   Lcurve%K_S%Prob = kolmogorov_smirnov_prob(sqrt(en**2.0/(2.0*en))*D)
   KSprob = Lcurve%K_S%Prob

  end subroutine kolmogorov_smirnov_test2

  !-------------------------------------------------------------------------------------
!  subroutine kolmogorov_smirnov_test1(Counts, BCounts, Nbin, Lcurve,&
!                                                Lcurve_bkg, D, KSprob)
  !-------------------------------------------------------------------------------------
  ! Compares the cumulative rate distribution with the time
  ! distribution. Returns the corresponding K_S statistic D, and the signifiance level
  ! Prob. Great values of Prob indicate that there is no variability in the input
  ! distribution.

!    real(kind=single),dimension(:) :: Counts, BCounts
!    integer,intent(in) :: Nbin
!    type(data_tests),intent(inout) :: Lcurve, Lcurve_bkg

!    real(kind=single) :: D, en, Nbin_Real, ir, Cumulative_Poisson_Value
!    real(kind=single) :: Cumulative_Poisson_Value_bkg, KSprob, fn, f0
!    integer :: i, j

!    D = 0.0; f0=0.0
!    call sort(Counts)
!    Nbin_Real = real(Nbin)
!    en=Nbin_Real
    
!!    en = (sqrt((Nbin_Real*Nbin_Real)/(Nbin_Real+Nbin_Real)))
!!    en = sqrt(Nbin_Real)
!!    ir = 0.
!    do j=1,Nbin_Real
!        fn=j/en  
!        ir=amax1(abs(f0-Counts(j)),abs(fn-Counts(j)))
!        if (ir > D) D=ir
!        f0=fn
!    end do
!   Lcurve%K_S%Stat = D
!   print*,sqrt(en)*D
!   Lcurve%K_S%Prob = kolmogorov_smirnov_prob(sqrt(en)*D)
!   KSprob = Lcurve%K_S%Prob
!!       Cumulative_Poisson_Value = 0.5*(1.0-Gauss_Integral_Function( &
!!            (Lcurve%Mean_Count-Counts(i)) / sqrt(2.0*Lcurve%Mean_Count)))
!!       Cumulative_Poisson_Value_bkg = 0.5*(1.0-Gauss_Integral_Function( &
!!         (Lcurve_bkg%Mean_Count_bkg-BCounts(i))/sqrt(2.0*Lcurve_bkg%Mean_Count_bkg)))
!!       D = max( D , abs(ir/Nbin_Real - Cumulative_Poisson_Value ) )
!!       ir = real(i)
!!       D = max( D , abs(ir/Nbin_Real - Cumulative_Poisson_Value) )
!!       D = max( D , abs(Cumulative_Poisson_Value_bkg - Cumulative_Poisson_Value) )
!!    end do
!!    Lcurve%K_S%Stat = D
!!    Lcurve%K_S%Prob = kolmogorov_smirnov_prob((en+0.12+0.11/en) * D)
!!    KSprob = Lcurve%K_S%Prob

!  end subroutine kolmogorov_smirnov_test1

  !-------------------------------------------------------------------------------------
  ! Many numerical functions or subroutines used by previous algorithms of variability
  ! are farther implemented. They might be icluded in a library if other SAS task want
  ! to use them (Gauss_Integral_Function, gamma_function_P, gamma_function_Q,
  ! series_gamma_func, fraction_gamma_function, gamma_log).
  !-------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------------
  subroutine median(Counts, Nbin, median_value)
  !-------------------------------------------------------------------------------------
  ! Returns the median.

    real(kind=single),dimension(:) :: Counts
    real(kind=single),allocatable,dimension(:) :: sortcounts
    integer,intent(in) :: Nbin
    real :: median_value
    integer :: middle

    allocate(sortcounts(Nbin))

    sortcounts=Counts
 
    call sort(sortcounts)
    middle = nint(real(Nbin)/2.0)
    median_value = sortcounts(middle)

    deallocate(sortcounts)

  end subroutine median 

  !-------------------------------------------------------------------------------------
  function Gauss_Integral_Function(x)
  !-------------------------------------------------------------------------------------
  ! Returns the error function Gauss_Integral_Function(x).

    real(kind=single),intent(in) :: x
    real(kind=single) :: Gauss_Integral_Function

    if (x < 0.) then
       Gauss_Integral_Function = -gamma_function_P(0.5 , x**2.0)
    else
       Gauss_Integral_Function = gamma_function_P(0.5 , x**2.0)
    end if
  end function Gauss_Integral_Function

  !-------------------------------------------------------------------------------------
  function gamma_function_P(a,x)
  !-------------------------------------------------------------------------------------
  ! Returns the incomplete gamma function P(a,x) for a > 0. and x >= 0.

    real(kind=single),intent(in) :: a , x
    real(kind=single) :: gamma_function_P , gammcf , gamser , gln

    if (x < a+1.0) then
       call series_gamma_func(gamser,a,x,gln)       
! Uses the series representation
       gamma_function_P = gamser
    else
       call Fraction_gamma_function(gammcf,a,x,gln) 
! Uses the complement of the
       gamma_function_P = 1.0 - gammcf               
! continued fraction representation
    end if
  end function gamma_function_P

  !-------------------------------------------------------------------------------------
  function gamma_function_Q(a,x)
  !-------------------------------------------------------------------------------------
  ! Returns the incomplete gamma function Q(a,x)=1-P(a,x) for a > 0. and x >= 0.

    real(kind=single),intent(in) :: a , x
    real(kind=single) :: gamma_function_Q , gammcf , gamser , gln

    if (x < a+1.0) then
       call series_gamma_func(gamser,a,x,gln)         
! Uses the complement of the series
       gamma_function_Q = 1.0 - gamser                 
! representation
    else
       call Fraction_gamma_function(gammcf,a,x,gln)   
! Uses the continued fraction
       gamma_function_Q = gammcf                      
! representation
    end if
  end function gamma_function_Q

  !-------------------------------------------------------------------------------------
  subroutine series_gamma_func(gamser ,a , x , gln)
  !-------------------------------------------------------------------------------------
  ! Returns the incomplete gamma function P(a,x) evaluated by its series representation
  ! as gamser. Also returns Ln(gamma(a) as gln. Warning : this routine works only if 
  ! a > 0. and x >= 0.

    real(kind=single),intent(in) :: a , x
    real(kind=single),intent(out) :: gamser , gln
    real(kind=single),parameter :: eps = 3.0e-7
    integer,parameter :: imax = 100
    integer :: i
    real(kind=single) :: del , sum

    gln = gamma_log(a)
    if ( x == 0. ) then
       gamser = 0.0
    else
       del = 1.0/a
       sum = del
       i = 1
       do while ( (i <= imax) .and. (abs(del) >= abs(sum)*eps) )
          del = del * x/(a+i)
          sum = sum + del
          i = i + 1
       end do
       gamser = sum * exp(-x + a*log(x) - gln)
    end if
  end subroutine series_gamma_func

  !-------------------------------------------------------------------------------------
  subroutine fraction_gamma_function(gammcf ,a , x , gln)
  !-------------------------------------------------------------------------------------
  ! Returns the incomplete gamma function Q(a,x) evaluated by its continued 
  ! fraction representation as gammcf. Also returns Ln(gamma(a)) as gln. 
  ! Warning : this routine works only if x >0. and a > 0.

    real(kind=single),intent(in) :: a , x
    real(kind=single),intent(out) :: gammcf , gln
    real(kind=single),parameter :: eps = 3.e-7 , min_float = 1.e-30
    integer,parameter :: imax = 100
    integer :: i
    real(kind=single) :: an , b , c , d , del , h

    gln = gamma_log(a)
    b = x + 1.0 - a
    c = 1./min_float
    d = 1./b
    h = d
    i = 1
    del = 2.0                     ! Non significative value use for first loop iteration.
    do while ( (i <= imax) .and. (abs(del-1.0) >= eps) )
       an = -i*(i-a)
       b = b + 2.0
       d = an*d + b
       if ( abs(d) < min_float ) d = min_float
       c = b + an/c
       if ( abs(c) < min_float ) c = min_float
       d = 1.0/d
       del = d * c
       h = h * del
    end do
    gammcf = exp(-x + a*log(x) - gln) * h
  end subroutine fraction_gamma_function

  !-------------------------------------------------------------------------------------
  function gamma_log(xx)
  !-------------------------------------------------------------------------------------
  ! Returns Lanczos approximation of the value Ln(gamma(xx)) for xx > 0.

    real(kind=single),intent(in) :: xx
    real(kind=single) :: gamma_log
    real(kind=single) :: ser
    real(kind=double),dimension(6) :: cof = (/ 76.18009172947146d0 , &
         -86.50532032941677d0 , 24.01409824083091d0 , -1.231739572450155d0 , &
         .1208650973866179d-2 , -0.5395239384953d-5 /)
    real(kind=single) :: stp = 2.5066282746310005d0
    real(kind=single) :: x , y , tmp
    integer :: j

    ser = 1.000000000190015
    x = xx
    y = x
    tmp = x + 5.5
    tmp = (x+0.5) * log(tmp) - tmp
    do j=1,6
       y = y + 1.0
       ser =  ser + cof(j)/y
    end do
    gamma_log = real(tmp + log(stp * ser/x))
  end function gamma_log

  !-------------------------------------------------------------------------------------
  recursive subroutine sort(arr)
  !-------------------------------------------------------------------------------------
  ! Sorts an array arr(1:n) into ascending numerical order. arr is replaced on output
  ! by its sorted rearrangement

    real(kind=single),dimension(:),intent(inout) :: arr
    integer :: begin , end
    real(kind=single) :: elt , tmp
    integer :: i , j , med

    begin = 1 ; end = size(arr)
    if (end > begin) then
       elt = arr(begin)
       i = begin + 1 ; j = end
       do while (i < j)
          do while ( (arr(i) < elt) .and. (i < j) )
             i = i + 1 ; arr(i) = arr(i)
          end do
          do while ( (arr(j) >= elt) .and. (j > i) )
             j = j - 1 ; arr(j) = arr(j)
          end do
          if (i < j) then
             tmp = arr(i)
             arr(i) = arr(j)
             arr(j) = tmp
          end if
       end do
       if (arr(i) < elt) then
          med = i
       else
          med = i - 1
       end if
       tmp = arr(begin)
       arr(begin) = arr(med)
       arr(med) = tmp
       if (med-begin > 1) call sort(arr(begin : med-1))
       if (end-med > 1) call sort(arr(med+1 : end))
    end if
  end subroutine sort

  !-------------------------------------------------------------------------------------
  function kolmogorov_smirnov_prob(lambda)
  !-------------------------------------------------------------------------------------
  ! Kolmogorov-Smirnov probability function.

!    real(single), intent(in) :: lambda

!    real(kind=single) :: kolmogorov_smirnov_prob
!    real(single) :: convCriterion, sign, summ, term
!    integer, parameter :: maxJ = 100
!    integer :: j

    !-------------------------------------------------------------------
    ! Start:

!    convCriterion = 10.0**(1.0 - precision(lambda))

!    j = 1
!    sign = 1.0
!    summ = 0.0
!    do
!      term = sign * exp(-2.0 * j * j * lambda * lambda)
!      summ = summ + term

!      if (abs(term) < convCriterion * summ)& ! converged
!    exit

!      j = j + 1
!      if (j >= maxJ) then ! routine hasn't converged, set ksQFunction = 1
!        kolmogorov_smirnov_prob = 0.0
!  return
!      end if

!      sign = -sign
 !   end do

!    kolmogorov_smirnov_prob = 2.0 * summ

    real(kind=single) :: kolmogorov_smirnov_prob
    real(kind=single),intent(in) :: lambda
    real(kind=single),parameter :: eps1=0.001 , eps2=1.e-8
    integer :: j
    real(kind=single) :: lambda2 , term , last_term , fac

    lambda2 = -2.*((0.1*lambda)**2.0)
    fac = 2.
    kolmogorov_smirnov_prob = 0.0
    last_term = 0.

     loopj: do j=1,99
     term = fac*exp(lambda2*j**2.0)
     kolmogorov_smirnov_prob = kolmogorov_smirnov_prob + term
     if ((abs(term) < eps1*last_term) .or. &
         (abs(term) < eps2*kolmogorov_smirnov_prob)) exit loopj
       last_term = abs(term)
       fac = -fac
     end do loopj

    if (kolmogorov_smirnov_prob > 1.0d0) kolmogorov_smirnov_prob =1.0d0
    if (j == 100) kolmogorov_smirnov_prob = 1.0d0

  end function kolmogorov_smirnov_prob

  !-------------------------------------------------------------------------------------
  function belong(String , String_Table)
  !-------------------------------------------------------------------------------------
  ! Returns a boolean telling if a string belongs to a string table.

    logical :: belong
    character(len=*),intent(in) :: String
    character(len=*),dimension(:),intent(in) :: String_Table
    integer :: i

    belong = .false.
    do i=1,size(String_Table)
       belong = belong .or. ( String == String_Table(i) )
    end do

  end function belong

  !-------------------------------------------------------------------------------------
  subroutine string_of_real(Nb , String , Length)
  !-------------------------------------------------------------------------------------
  ! Converts real numbers into plottable character chain. Gives also its length.

    real(kind=single),intent(in) :: Nb
    character(len=20),intent(out) :: String
    integer,intent(out) :: Length
    integer :: MM , puiss

!    puiss = max( 0 , min( 6 , 6-int(log10(Nb)) ) )
    puiss = max( 0 , min( 8 , 8-int(log10(Nb)) ) )
    MM = nint(Nb*(10**puiss))
    call pgnumb(MM , -puiss , 0 , String , Length)

  end subroutine string_of_real

  !-------------------------------------------------------------------------------------
  subroutine string_of_real2(Nb , String , Length)
  !-------------------------------------------------------------------------------------
  ! Converts real numbers into plottable character chain. Gives also its length.

    real(kind=double),intent(in) :: Nb
    character(len=20),intent(out) :: String
    integer,intent(out) :: Length
    integer :: MM , puiss

    puiss = max( 0 , min( 8 , 8-int(log10(Nb)) ) )
    MM = nint(Nb*(10**puiss))
    call pgnumb(MM , -puiss , 0 , String , Length)

  end subroutine string_of_real2

  !-------------------------------------------------------------------------------------
  function change_case(instring,dir) result(outstring)
  !-------------------------------------------------------------------------------------
  ! Switch from lower to upper case or vice-versa: From Jean Ballet
  ! instring : input string
  ! dir      : direction (1: to lower case; 2: to uppercase; 0: switch case)

    character(len=*), intent(in) :: instring
    integer, intent(in) :: dir
    character(len=len(instring)) :: outstring

  ! ASCII positions of lower case and upper case letters

    integer, parameter :: start_up  = iachar("A"), end_up  = iachar("Z"), &
                          start_low = iachar("a"), end_low = iachar("z"), &
                          offset = start_low - start_up

    integer :: length, icur, pos

    length = len(instring)
    outstring = instring
    do icur = 1, length
       pos = iachar(instring(icur:icur))
       if (pos >= start_up .and. pos <= end_up .and. dir < 2) then
          pos = pos + offset
       else if (pos >= start_low .and. pos <= end_low .and. dir /= 1) then
          pos = pos - offset
       endif
       outstring(icur:icur) = achar(pos)
    enddo

  end function change_case  
  
end module ekstest_mod
